/*
* File        ReportDownload.c
* Usage       ReportDownload <inifile> [optional businessdate YYYYMMDD]
* Description Sample program to query and download a list of participant reports from OMnet.
* Version     1.2
* License     BSD
*
Copyright (c) 2013, Australian Securities Exchange (ASX)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met: 

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer. 
2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <e4c.h>

#include <om_inttypes.h>
#include <omnifact.h>
#include <omni.h>
#include <omex_sfe.h>
#include <omniapi.h>

/*----------------------------- Global Variables/Macros ------------------------------*/
#define ASCII 1
#define BINARY 2
#define MAX_REPORT_LIST 100
#define MAX_NEWPWDLEN 20

E4C_DECLARE_EXCEPTION(FileException);
E4C_DECLARE_EXCEPTION(FileOpenException);
E4C_DECLARE_EXCEPTION(FileReadException);
E4C_DECLARE_EXCEPTION(FileNotFoundException);
E4C_DECLARE_EXCEPTION(MaxReportListException);
E4C_DECLARE_EXCEPTION(OmexException);
E4C_DECLARE_EXCEPTION(CreateSessionException);
E4C_DECLARE_EXCEPTION(GetFacilityTypeException);
E4C_DECLARE_EXCEPTION(ReadIniFileException);
E4C_DECLARE_EXCEPTION(OmnetLoginException);
E4C_DECLARE_EXCEPTION(SendTxException);
E4C_DECLARE_EXCEPTION(InvalidFacilityTypeException);
E4C_DECLARE_EXCEPTION(GetUserInfoException);
E4C_DECLARE_EXCEPTION(UpdateIniFileException);
E4C_DECLARE_EXCEPTION(GetBusinessDateException);
E4C_DECLARE_EXCEPTION(QueryException);
E4C_DECLARE_EXCEPTION(GetReportFilesException);
E4C_DECLARE_EXCEPTION(GetReportDataException);
E4C_DECLARE_EXCEPTION(InvalidDateFormatException);
E4C_DECLARE_EXCEPTION(ExceededBufferSizeException);

E4C_DEFINE_EXCEPTION(FileException, "File access exception.", RuntimeException);
E4C_DEFINE_EXCEPTION(FileOpenException, "Failed to open file.", FileException);
E4C_DEFINE_EXCEPTION(FileReadException, "Failed to read file.", FileException);
E4C_DEFINE_EXCEPTION(FileNotFoundException, "File not found.", FileException);
E4C_DEFINE_EXCEPTION(MaxReportListException, "Exceeded max Report List.", RuntimeException);
E4C_DEFINE_EXCEPTION(OmexException, "Omex API call failure.", RuntimeException);
E4C_DEFINE_EXCEPTION(CreateSessionException, "Failed to create session.", RuntimeException);
E4C_DEFINE_EXCEPTION(GetFacilityTypeException, "Failed to get facility type.", RuntimeException);
E4C_DEFINE_EXCEPTION(ReadIniFileException, "Error reading ini file.", RuntimeException);
E4C_DEFINE_EXCEPTION(OmnetLoginException, "OMnet login failure.", RuntimeException);
E4C_DEFINE_EXCEPTION(SendTxException, "OMnet send transaction failure.", RuntimeException);
E4C_DEFINE_EXCEPTION(InvalidFacilityTypeException, "Facility type must not be 0.", RuntimeException);
E4C_DEFINE_EXCEPTION(GetUserInfoException, "Failed to get user information.", RuntimeException);
E4C_DEFINE_EXCEPTION(UpdateIniFileException, "Error updating ini file.", RuntimeException);
E4C_DEFINE_EXCEPTION(GetBusinessDateException, "Failed to get business date.", RuntimeException);
E4C_DEFINE_EXCEPTION(QueryException, "OMnet query failure.", RuntimeException);
E4C_DEFINE_EXCEPTION(GetReportFilesException, "Failed to get list of reports.", RuntimeException);
E4C_DEFINE_EXCEPTION(GetReportDataException, "Failed to download report.", RuntimeException);
E4C_DEFINE_EXCEPTION(InvalidDateFormatException, "Business date must be YYMMDD.", RuntimeException);
E4C_DEFINE_EXCEPTION(ExceededBufferSizeException, "Exceeded buffer size.", RuntimeException);

struct Login_t      //login info for OMnet
{
  char Username[32];
  char Password[32];
  char Host[16];
  char Port[8];
  int isLoggedIn;
};
struct Report_t     //report data
{
  int32_t ReportNumber;
  char ReportFile[40];
  uint8_t ReportType;
  char ReportName[80];
};

/* Session Handler */
static omniapi_session_handle hSession;

/* omniapi_tx_ex return values */
int32_t     glTransactionStatus;
uint32_t    glTransactionId[2];
uint32_t    glOrderId;

static uint32_t glFacilityType_EP0;

int depth=0;    //function call depth indicator;

/*----------------------------- Global Functions ------------------------------*/

void CheckError(int32_t errorMsgCode,int32_t TransactionStatus)
/*--------------------------------------------------------------------
Retrieve error messages associated with completion and transaction status.
Throws exception if error found.
---------------------------------------------------------------------*/
{
  int32_t completionStatus;
  char csmsg[256],tsmsg[256];
  uint32 cslen,tslen;

  csmsg[0]='\0';
  cslen=sizeof(csmsg);
  tsmsg[0]='\0';
  tslen=sizeof(tsmsg);
  if (errorMsgCode < 0)
  {
    completionStatus = omniapi_get_message_ex( hSession, errorMsgCode, csmsg, &cslen, 0);
    if( completionStatus < 0 )
    {
      sprintf(csmsg,"Failed to retrieve message for cstatus %d",errorMsgCode);
    }
    switch (TransactionStatus)
    {
	case 0:
		break;
    case LM_FILENOTFOUND:
      throw(FileNotFoundException,NULL);
      break;
    default:
      completionStatus = omniapi_get_message_ex( hSession, TransactionStatus, tsmsg, &tslen, 0);
      if( completionStatus <0)
      {
        sprintf(tsmsg,"Failed to retrieve message for transaction error %d",TransactionStatus);
      }
      strcat(csmsg,", ");
      strcat(csmsg,tsmsg);
      break;
    }
	throw(OmexException,csmsg);
  }
}

void strzcpy(char *dest,size_t dlen,char *source, size_t slen)
/*--------------------------------------------------------------------
copy source buffer to dest buffer. Insert string terminator(\0) at slen 
chars or when 1st non-space is encountered scanning from back to front.
---------------------------------------------------------------------*/
{
  int32_t blen;
  char *pdest,*psrc,*start,*endaddr;

  pdest=dest;
  psrc=source;
  blen=slen;
  endaddr=dest+dlen-1;

  if (dlen > slen)
  {
    strncpy(pdest,psrc,slen);
    start=pdest + slen-1;
  }
  else
  {
    strncpy(pdest,psrc,dlen);
    start=pdest + dlen-1;
  }
  for (pdest=start; pdest >= dest; pdest--)
  {
    if (*pdest != ' ')
    {
      if (pdest < endaddr )
        *(++pdest)='\0';
      else
        *pdest='\0';
      break;
    }
  }
  if (pdest < dest)
  {
    *(dest)='\0';
  }
}

void CreateSession()
/*--------------------------------------------------------------------
create new omnet session.
---------------------------------------------------------------------*/
{
  depth++;
  try
  {
    hSession = omniapi_create_session();
    printf("Create OMnet session.\n");
  }
  catch (RuntimeException)
  {
    const e4c_exception * exception = e4c_get_exception();
    printf("%d. CreateSession: (%s)%s\n",depth--,exception->name,exception->message);
    throw(CreateSessionException,NULL);
  }
  finally
  {
    depth--;
  }
}

void ClearSessionHandler()
/*--------------------------------------------------------------------
close session.
---------------------------------------------------------------------*/
{
  omniapi_close_session(hSession);
}

void GetFacilityType()
/*--------------------------------------------------------------------
get base facility type EP0.
---------------------------------------------------------------------*/
{
  uint32_t factypeSize;
  int32_t  completionStatus;
  depth++;

  printf("Retrieve facility type.\n");
  try
  {
    factypeSize = sizeof(glFacilityType_EP0);
    completionStatus = omniapi_get_info_ex(
      hSession,
      &glTransactionStatus, 
      OMNI_INFTYP_FACTYP_E0, 
      &factypeSize, 
      &glFacilityType_EP0);

    PUTLONG(glFacilityType_EP0,glFacilityType_EP0);
    CheckError(completionStatus,glTransactionStatus);
  }
  catch (RuntimeException)
  {
    const e4c_exception * exception = e4c_get_exception();
    printf("%d. GetFacilityType: (%s)%s\n",depth,exception->name,exception->message);
    throw(GetFacilityTypeException,NULL);
  }
  finally
  {
    depth--;
  }
}

void GetUserInfo(user_code_t *userData,int32_t userDataSize)
/*--------------------------------------------------------------------
get logged in user information
---------------------------------------------------------------------*/
{
  int32_t completionStatus;
  depth++;

  printf("Retrieve user information.\n");
  try
  {
    completionStatus = omniapi_get_info_ex(
      hSession,
      &glTransactionStatus, 
      OMNI_INFTYP_USERCODE, 
      &userDataSize, 
      userData);
    CheckError(completionStatus,glTransactionStatus);
  }
  catch (RuntimeException)
  {
    const e4c_exception * exception = e4c_get_exception();
    printf("%d. GetUserInfo: (%s)%s\n",depth,exception->name,exception->message);
    throw(GetUserInfoException,NULL);
  }
  finally
  {
    depth--;
  }

}

int32_t SendTx( void *buffer, uint32_t length, uint32_t facilityType )
/*--------------------------------------------------------------------
wrapper for omniapi_tx_ex (Issue an OMnet Transaction)
---------------------------------------------------------------------*/
{
  static omni_message *message[2]; /* only use the first item */
  static char sendBuffer[MAX_REQUEST_SIZE];
  int32_t  completionStatus;
  depth++;

  message[0] = (omni_message*)sendBuffer;
  message[1] = NULL;
  memcpy( &sendBuffer[0], &length, sizeof(length) );
  memcpy( &sendBuffer[sizeof(length)], buffer, length );

  try
  {
    if (facilityType == 0)
    {
      throw(InvalidFacilityTypeException,NULL);
    }

    completionStatus = omniapi_tx_ex(hSession, &glTransactionStatus, facilityType,
      message, &glTransactionId[0],&glOrderId );
    if (completionStatus != OMNIAPI_PASSW_EXPIRED)   //ignore password expired
    {
      CheckError(completionStatus,glTransactionStatus);
    }
  }
  catch (RuntimeException)
  {
    const e4c_exception * exception = e4c_get_exception();
    printf("%d. SendTx: (%s)%s\n",depth,exception->name,exception->message);
    throw(SendTxException,NULL);
  }
  finally
  {
    depth--;
  }
  return completionStatus;
}

void Query(void *qryBuffer, uint32_t qryLength, uint32_t facilityType,void *rcvBuffer, uint32_t *rcvLength )
/*--------------------------------------------------------------------
wrapper for omniapi_query_ex (Issue an OMnet Query)
---------------------------------------------------------------------*/
{
  static char sendBuffer[MAX_REQUEST_SIZE];
  static omni_message *queryMessage;
  quad_word dummyOrderId;
  int32_t completionStatus;
  depth++;

  try
  {
    queryMessage = (omni_message*)sendBuffer;
    memcpy( &sendBuffer[0], &qryLength, sizeof(qryLength) );
    memcpy( &sendBuffer[sizeof(qryLength)], qryBuffer, qryLength );
    completionStatus = omniapi_query_ex(
      hSession,
      &glTransactionStatus,
      facilityType,
      queryMessage,
      1,            /* TRUE=return status */
      (int8 *)rcvBuffer,
      rcvLength,
      &glTransactionId[0],
      (uint32_t *)&dummyOrderId );
    CheckError(completionStatus,glTransactionStatus);
  }
  catch (FileNotFoundException)
  {
    rethrow("");
  }
  catch (RuntimeException)
  {
    const e4c_exception * exception = e4c_get_exception();
    printf("%d. Query: (%s)%s\n",depth,exception->name,exception->message);
    throw(QueryException,NULL);
  }
  finally
  {
    depth--;
  }
}

void RandomPassword(char newpwd[],int32_t newpwdlen)
/*--------------------------------------------------------------------
Generate a new password with random letters and numbers - xxxx_nnnn
---------------------------------------------------------------------*/
{
  int32_t rannum;

  newpwd[0]='\0';
  if (newpwdlen >= MAX_NEWPWDLEN)
  {
    srand((int)time(NULL));
    newpwd[0]=rand() % ('Z'-'A')+ 1 + 'A';    //A-Z
    newpwd[1]=rand() % ('z'-'a')+ 1 + 'a';    //a-z
    newpwd[2]=rand() % ('Z'-'A')+ 1 + 'A';    //A-Z
    newpwd[3]=rand() % ('z'-'a')+ 1 + 'a';    //a-z
    newpwd[4]='_';
    rannum=rand() % 10000;    //0-9999
    sprintf(&newpwd[5],"%04d",rannum);
  }
}

void UpdateIniFile(char *InputFile,int32_t lineno,char *record)
/*--------------------------------------------------------------------
Replace record# <lineno> in a text file with a new line.
---------------------------------------------------------------------*/
{
  int32_t count;
  FILE *infp=NULL,*outfp=NULL;
  char tmpfile[L_tmpnam];
  char buf[256];
  depth++;

  tmpnam(tmpfile);    //get temp file name
  try
  {
    if ((infp=fopen(InputFile, "r")) == NULL) throw(FileOpenException,InputFile);
    if ((outfp=fopen(tmpfile, "w")) == NULL) throw(FileOpenException,tmpfile);
    for (count=1; count < lineno; count++)  //keep lines before lineno
    {
      fgets(buf,sizeof(buf),infp);    //read from original file
      fputs(buf,outfp);               //write to temp file
    }
    fprintf(outfp,"%s\n",record);     //write new line
    fgets(buf,sizeof(buf),infp);      //skip old line
    while (fgets(buf, sizeof(buf), infp) != NULL)   //keep the rest
    {
      fputs(buf,outfp);
    }
    fclose(infp);
    fclose(outfp);
    remove(InputFile);          //delete original file
    rename(tmpfile,InputFile);  //rename temp file to original file
  }
  catch (RuntimeException)
  {
    const e4c_exception * exception = e4c_get_exception();
    printf("%d. UpdateIniFile: (%s)%s\n",depth,exception->name,exception->message);
    throw(UpdateIniFileException,NULL);
  }
  finally
  {
    if (infp != NULL) fclose(infp);
    if (outfp != NULL) fclose(outfp);
    depth--;
  }
}

void OmnetLogin(struct Login_t *Login,char *inifile)
/*--------------------------------------------------------------------
perform Omnet login using login data. Handle password expiry by
generating new random password and updating inifile.
---------------------------------------------------------------------*/
{
  char newpwd[MAX_NEWPWDLEN];
  char errmsg[50];
  omni_login_t loginMessage;
  omni_set_password_t setNewpwd;
  int32_t completionStatus;
  int32_t txStatus;
  depth++;

  Login->isLoggedIn=0;     //default not logged in
  memset( &loginMessage, 0, sizeof(loginMessage) );  
  strncpy(loginMessage.user_s, Login->Username,    sizeof(loginMessage.user_s) );
  strncpy(loginMessage.pass_s, Login->Password,    sizeof(loginMessage.pass_s) );
  strncpy(loginMessage.gateway_node_s, Login->Host,        sizeof(loginMessage.gateway_node_s) );
  loginMessage.port_u=atoi(Login->Port);
  strncpy(loginMessage.appl_ident_s, "ReportDownload", sizeof(loginMessage.appl_ident_s) );
  loginMessage.forced_u = LOGIN_NORMAL;
  printf("Login username: %s\n",Login->Username);
  printf("Login Host: %s\n",Login->Host);
  printf("Login Port: %s\n",Login->Port);

  try
  {
	  completionStatus = omniapi_login_ex( hSession, &txStatus, &loginMessage );
    if (completionStatus == OMNIAPI_PWD_CHANGE_REQ)   //password expired?
    {
	    memset( &setNewpwd, 0, sizeof(setNewpwd) );
      RandomPassword(newpwd,sizeof(newpwd));      //new random password
      strncpy(loginMessage.pass_s, newpwd, sizeof(loginMessage.pass_s));
      strncpy(setNewpwd.pass_s, Login->Password, sizeof(setNewpwd.pass_s));
	    strncpy(setNewpwd.new_pass_s, newpwd, sizeof(setNewpwd.new_pass_s));
	    completionStatus = omniapi_set_newpwd_ex( hSession, &txStatus, &setNewpwd);
      UpdateIniFile(inifile,2,newpwd);    //update ini file
    }
	  else if (completionStatus >= 0)
	  {
	    Login->isLoggedIn=1;   //set flag true
	  }
	  else
	  {
      sprintf(errmsg,"omniapi_login_ex call failed, status = %d", completionStatus);
      throw(RuntimeException,errmsg);
	  }
  }
  catch (RuntimeException)
  {
    const e4c_exception * exception = e4c_get_exception();
    printf("%d. OmnetLogin: (%s)%s\n",depth,exception->name,exception->message);
    throw(OmnetLoginException,NULL);
  }
  finally
  {
    depth--;
  }
}

int32_t Logout()
/*--------------------------------------------------------------------
perform OMnet logout
---------------------------------------------------------------------*/
{
  int32_t  completionStatus,txStatus;
  depth++;

  printf("Logout from OMnet.\n");
  try
  {
	  completionStatus = omniapi_logout_ex ( hSession, &txStatus );
  }
  finally
  {
    depth--;
  }
  return completionStatus;
}

void GetBusinessDate(char *prevdate,int32_t buflen)
/*--------------------------------------------------------------------
get previous business date from the first product returned.
---------------------------------------------------------------------*/
{
  query_clearing_date_t queryBuffer;
  answer_clearing_date_t answerBuffer;
  uint32_t    answerLength;
  int32_t count=0,total=0;
  uint32_t FacilityType;
  depth++;

  FacilityType=glFacilityType_EP0+5;
  memset(&queryBuffer, 0, sizeof(queryBuffer));
  queryBuffer.transaction_type.central_module_c     = 'C';
  queryBuffer.transaction_type.server_type_c        = 'Q';
  queryBuffer.transaction_type.transaction_number_n =  68;

  PUTSHORT(queryBuffer.transaction_type.transaction_number_n, queryBuffer.transaction_type.transaction_number_n);

  *prevdate='\0';

  try
  {
    answerLength = sizeof(answerBuffer);
    Query(&queryBuffer, sizeof(queryBuffer), FacilityType, &answerBuffer, &answerLength );
    count=0;    //get prev business date from first item.
    strzcpy(prevdate,buflen,(char *)answerBuffer.item[count].prev_clearing_date_s,sizeof(answerBuffer.item[count].prev_clearing_date_s));
  }
  catch (RuntimeException)
  {
    const e4c_exception * exception = e4c_get_exception();
    printf("%d. GetBusinessDate: (%s)%s\n",depth,exception->name,exception->message);
    throw(GetBusinessDateException,NULL);
  }
  finally
  {
    depth--;
  }
}

void GetReportData(struct Report_t ReportList,char *busdate,int *total)
//--------------------------------------------------------------------
//This method calls LQ3 with a given report number and business date to
//extract report data from server.
//--------------------------------------------------------------------
{
  static query_list_ver_t queryBuffer;
  static answer_list_ver_t answerBuffer;
  uint32_t    answerLength;
  int32_t count;
  uint32_t glFacilityType_EP4;
  static char textBuffer[1000];
  char list_name_s[40],file_type_s[8];
  char *p;
  uint16_t blen;
  FILE *fp=NULL;
  depth++;

  glFacilityType_EP4=glFacilityType_EP0+4;

  memset(&queryBuffer, 0, sizeof(queryBuffer));
  queryBuffer.transaction_type.central_module_c     = 'L';
  queryBuffer.transaction_type.server_type_c        = 'Q';
  queryBuffer.transaction_type.transaction_number_n =  3;

  strcpy(queryBuffer.date_s, busdate)  ;     //business date
  queryBuffer.info_type_i = ReportList.ReportNumber;            //report number from an earlier LQ4 call.

  PUTSHORT(queryBuffer.transaction_type.transaction_number_n, queryBuffer.transaction_type.transaction_number_n);

  answerBuffer.segment_number_n = 1;     //this is used as a flag to see if we are done
  queryBuffer.segment_number_n = 0;

  try
  {
    while ( answerBuffer.segment_number_n != 0 )
    {
      queryBuffer.segment_number_n++;
      PUTSHORT(queryBuffer.segment_number_n, queryBuffer.segment_number_n);
      answerLength = sizeof(answerBuffer);

	  //printf("Before call %d\n",answerBuffer.segment_number_n);
      Query(&queryBuffer, sizeof(queryBuffer), glFacilityType_EP4, &answerBuffer, &answerLength );
	  //printf("After call %d\n",answerBuffer.segment_number_n);

      if (queryBuffer.segment_number_n == 1)    //first time in loop?
      {
        strzcpy(list_name_s,sizeof(list_name_s),(char *)answerBuffer.list_name_s,sizeof(answerBuffer.list_name_s));
        strzcpy(file_type_s,sizeof(file_type_s),(char *)answerBuffer.file_type_s,sizeof(answerBuffer.file_type_s));
        strcat(list_name_s,".");
        strcat(list_name_s,file_type_s);
        printf("Saving to %s\n\n",list_name_s);
        if (ReportList.ReportType == ASCII)
          fp=fopen(list_name_s,"w");
        else
          fp=fopen(list_name_s,"wb");
      } 
      p = (char *)&answerBuffer.text_buffer_s;
      for (count=0; count < answerBuffer.items_n; count++)
      {
        memcpy( &blen, p, sizeof(blen) );
        p += sizeof(blen);
        if (ReportList.ReportType == ASCII)
        {
          if (blen+1 >= sizeof(textBuffer))
            throw(ExceededBufferSizeException,NULL);
          memcpy( &textBuffer, p, blen );
          textBuffer[blen]='\n';
          textBuffer[blen+1]='\0';
          //printf("%s",textBuffer);
          fputs(textBuffer,fp);
        }
        else
        {
          fwrite(p, 1, blen, fp);
        }
        if (blen % 2 > 0)       //odd-length strings are zero-padded,
          p = p + blen + 1;     //so must skip 1 byte
        else
          p += blen;            //even-length buffer are ok
      }
    }
    (*total)++;
  }
  catch (FileNotFoundException)
  {
    printf("File not found.\n\n");
  }
  catch (RuntimeException)
  {
    const e4c_exception * exception = e4c_get_exception();
    printf("%d. GetReportData: (%s)%s\n",depth,exception->name,exception->message);
    throw(GetReportDataException,NULL);
  }
  finally
  {
    if (fp != NULL) fclose(fp);
    depth--;
  }
}

void GetReportFiles(struct Report_t ReportList[],int32_t ReportListCount,char *busdate)
/*--------------------------------------------------------------------
This method calls LQ4 with infotype=256 to extract list of all report numbers and names
available in server.
--------------------------------------------------------------------*/
{
  query_report_ver_t queryBuffer;
  static answer_report_ver_t answerBuffer;
  static answer_report_ver_item_t  reportItem;
  uint32_t    answerLength;
  int32_t     completionStatus=0;
  int32_t count,index,total=0;
  uint32_t glFacilityType_EP4;
  struct Report_t ReportInfo;
  depth++;

  printf("*** Download reports for business date of %s ****\n",busdate);

  glFacilityType_EP4=glFacilityType_EP0+4;

  memset(&queryBuffer, 0, sizeof(queryBuffer));
  memset(&ReportInfo, 0, sizeof(ReportInfo));

  queryBuffer.transaction_type.central_module_c     = 'L';
  queryBuffer.transaction_type.server_type_c        = 'Q';
  queryBuffer.transaction_type.transaction_number_n =  4;

  queryBuffer.info_type_i = 256;    //get all report# and names.

  PUTSHORT(queryBuffer.transaction_type.transaction_number_n, queryBuffer.transaction_type.transaction_number_n);

  answerBuffer.segment_number_n = 1; /* this is used as a flag to see if we are done */

  queryBuffer.segment_number_n = 0;

  try
  {
    while ( answerBuffer.segment_number_n != 0 )
    {
      queryBuffer.segment_number_n++;
      PUTSHORT(queryBuffer.segment_number_n, queryBuffer.segment_number_n);
      answerLength = sizeof(answerBuffer);
      Query(&queryBuffer, sizeof(queryBuffer), glFacilityType_EP4, &answerBuffer, &answerLength );
      for (count=0; count < answerBuffer.items_n; count++)
      {
		    strzcpy(ReportInfo.ReportFile,sizeof(ReportInfo.ReportFile),answerBuffer.item[count].name_s,sizeof(answerBuffer.item[count].name_s));
		    strzcpy(ReportInfo.ReportName,sizeof(ReportInfo.ReportName),answerBuffer.item[count].description_s,sizeof(answerBuffer.item[count].description_s));
		    ReportInfo.ReportNumber=answerBuffer.item[count].info_type_i;
        ReportInfo.ReportType=answerBuffer.item[count].ascii_bin_c;

        //compare against list of required reports.
        for (index=0; index < ReportListCount; index++)   
        {
          if (strcmp(ReportInfo.ReportName,ReportList[index].ReportName) == 0 )   //if found,
          {
            ReportList[index].ReportNumber=ReportInfo.ReportNumber;   //get report number
            strcpy(ReportList[index].ReportFile,ReportInfo.ReportFile);   //report file name
            switch (ReportInfo.ReportType) {
            case 1:   //ascii
            case 4:   //html
            case 7:   //asc+tab
            case 8:   //csv
              ReportList[index].ReportType=ASCII;
              break;
            default:
              ReportList[index].ReportType=BINARY;
              break;
            }

          }
        }

      }
    }

    total=0;
    //process list of required reports
    for (index=0; index < ReportListCount; index++) 
    {
      printf("Requesting %s.\n",ReportList[index].ReportName);
      if (ReportList[index].ReportNumber != 0 )   //if report is available
      {
        GetReportData(ReportList[index],busdate,&total);   //download report
      }
      else
        printf("Report not available.\n\n");
    }
    printf("**** Total %d out of %d reports ****\n",total,ReportListCount); 
  }
  catch (RuntimeException)
  {
    const e4c_exception * exception = e4c_get_exception();
    printf("%d. GetReportFiles: (%s)%s\n",depth,exception->name,exception->message);
    throw(GetReportFilesException,NULL);
  }
  finally
  {
    depth--;
  }
}

int32_t ReadIniFile(struct Login_t *Login, struct Report_t ReportList[], char *InputFile)
/*--------------------------------------------------------------------
read ini file and assign data to fields in various structures. Returns count of required reports.
Ini file consists of:
Username
Password
Host
Port
ReportName1
Reportname2
:
<max MAX_REPORT_LIST reports>
--------------------------------------------------------------------*/
{
  char buf[100];
  int32_t index=0,buflen;
  FILE *fp=NULL;
  depth++;

  try
  {
    if ((fp=fopen(InputFile, "r")) == NULL) throw(FileOpenException,InputFile);
    if (fgets(Login->Username,sizeof(Login->Username),fp) != NULL)
    {
      buflen=strlen(Login->Username);
      if (buflen > 0)
        *(Login->Username+buflen-1) = '\0';
    }
    if (fgets(Login->Password,sizeof(Login->Password),fp) != NULL)
    {
      buflen=strlen(Login->Password);
      if (buflen > 0)
        *(Login->Password+buflen-1) = '\0';
    }
    if (fgets(Login->Host,sizeof(Login->Host),fp) != NULL)
    {
      buflen=strlen(Login->Host);
      if (buflen > 0)
        *(Login->Host+buflen-1) = '\0';
    }
    if (fgets(Login->Port,sizeof(Login->Port),fp) != NULL)
    {
      buflen=strlen(Login->Port);
      if (buflen > 0)
        *(Login->Port+buflen-1) = '\0';
    }
    while (fgets(buf, sizeof(buf), fp) != NULL)
    {
      buflen=strlen(buf);
      if (buflen > 0)
      {
        *(buf+buflen-1)='\0';
        if (strlen(buf) > 0)
        {
          strcpy(ReportList[index].ReportName,buf);
          ReportList[index].ReportNumber=0;
          index++;
          if (index == MAX_REPORT_LIST) throw(MaxReportListException,NULL);
        }
      }
    }
  }
  catch (RuntimeException)
  {
    const e4c_exception * exception = e4c_get_exception();
    printf("%d. ReadIniFile: (%s)%s\n",depth,exception->name,exception->message);
    throw(ReadIniFileException,NULL);
  }
  finally
  {
    if (fp != NULL) fclose(fp);
    depth--;
  }
  return index;
}

/*----------------------------- Main Function ------------------------------*/

int main(int argc, char *argv[])
/*
argv[0]  exe path name
argv[1]  inifile
argv[2]  YYYYMMDD (optional business date)
- read ini file and get list of required reports.
- Get list of all report names from OMnet using LQ4.
- Compare report name against required list and get report number/type.
- Download required reports by calling LQ3 with report number.
*/
{
  struct Login_t Login;
  struct Report_t ReportList[MAX_REPORT_LIST];
  int32_t ReportListCount;
  user_code_t userData;
  char inifile[256];
  char prevdate[9];
  time_t rawtime;

  if( argc < 2 )
  {
    printf("Usage: ReportDownload <inifile> [optional businessdate YYYYMMDD]\n");
    return 1;
  }
  if (strlen(argv[1])+1 <= sizeof(inifile))
  {
    strcpy(inifile,argv[1]);
  }
  else
  {
    printf("inifile name is too long.\n");
    return 1;
  }
  if (argc > 2)    //got business date?
  {
    if (strlen(argv[2])+1 != sizeof(prevdate))
    {
      printf("Business date format must be YYYYMMDD.\n");
      return 1;
    }
    else
    {
      strcpy(prevdate,argv[2]); //copy bus date
    }
  }

  memset(&Login, 0, sizeof(Login));
  memset(&ReportList, 0, sizeof(ReportList));

  Login.isLoggedIn=0;
  e4c_using_context(E4C_TRUE, e4c_print_exception)    //print all uncaught exceptions
  {
    depth++;
    try
    {
      time ( &rawtime );
      printf ( "\nDate of run: %s", ctime (&rawtime) );

      CreateSession();

      ReportListCount=ReadIniFile(&Login,ReportList,inifile);

      OmnetLogin(&Login,inifile);

      GetFacilityType();

      GetUserInfo(&userData,sizeof(userData));

      if (argc < 3)    //no business date?
      {
        GetBusinessDate(prevdate,sizeof(prevdate));   //use prev business date
      }

      GetReportFiles(ReportList,ReportListCount,prevdate);

      Logout();

      ClearSessionHandler();

    }
    catch (RuntimeException)
    {
      const e4c_exception * exception = e4c_get_exception();
      printf("%d. main: (%s)%s\n", depth,exception->name, exception->message); 
      if (Login.isLoggedIn)
      {
        Logout();
        ClearSessionHandler();
      }
    }

  }
}
